{#-
 # Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
 # Copyright (C) 2020  UAVCAN Development Team  <uavcan.org>
 # This software is distributed under the terms of the MIT License.
 # Authors: David Lenfesty, Scott Dixon <dixonsco@amazon.com>, Pavel Kirienko <pavel@uavcan.org>,
 #          Peter van der Perk <peter.vanderperk@nxp.com>
-#}

{% extends "base.j2" %}
{% block contents %}

{% from 'definitions.j2' import generate_metadata, generate_composite %}

{{ generate_metadata(T) }}

{{ generate_composite(T.request_type) }}

{{ generate_composite(T.response_type) }}

{% endblock %}
