#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# Copyright (C) 2018-2019  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#
"""
Contains the Jinja templates to generate C++ headers.
"""

__version__ = "1.0.0"
"""Version of the c++ template set."""
